using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using Utils.NetworkThread.Exceptions;

namespace Utils.NetworkThread
{
    /// <summary>
    /// Thread used for network communication
    /// </summary>
    public class ListenerNetworkThread
    {
        #region Variables
        /// <summary>
        /// Ip of the server
        /// </summary>
        private IPAddress vServerIP = null;
        /// <summary>
        /// Port of the server
        /// </summary>
        private int vServerPort = -1;
        /// <summary>
        /// Message received from the client
        /// </summary>
        private object vMessage = new object();
        /// <summary>
        /// Indicates whether the server has got continue listening
        /// </summary>
        private bool vContinueListen = true;
        /// <summary>
        /// Thread for the listening
        /// </summary>
        private Thread vThread;
        /// <summary>
        /// Socket Client
        /// </summary>
        private Socket vClient;
        /// <summary>
        /// Socjet Server
        /// </summary>
        private Socket vServer;
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public ListenerNetworkThread() { }
        /// <summary>
        /// Gets an instance of the thread
        /// </summary>
        /// <param name="_IP">IP of the server</param>
        /// <param name="_Port">Listening Port of the server</param>
        public ListenerNetworkThread(IPAddress _IP, int _Port)
        {
            vServerIP = _IP;
            vServerPort = _Port;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Kills asap the listening
        /// </summary>
        /// <exception cref="NetworkThreadException">Throws if problems occured when abording the thread</exception>
        public void Stop()
        {
            try
            {
                vContinueListen = false;
                if (vThread != null)
                    vThread.Abort();
                if (vClient != null)
                {
                    vClient.Shutdown(SocketShutdown.Both);
                    vClient.Close();                    
                }
                if (vServer != null)
                {                    
                    vServer.Close();
                }
            }
            catch (Exception ex)
            {
                throw new NetworkThreadException(ex.Message);
            }
        }
        /// <summary>
        /// Start listening
        /// </summary>
        /// <exception cref="NetworkThreadException">Throws if problems occured when starting the thread</exception>
        public void Start()
        {
            try
            {
                #region Test information
                if (vServerIP == null || vServerPort < 0)
                    throw new DataUnassignedException("The server IP and/or the server port are not assigned!");
                #endregion

                vThread = new Thread(new ThreadStart(Listen));
                vThread.Start();
            }
            catch (Exception ex)
            {
                throw new NetworkThreadException(ex.Message);
            }
        }
        /// <summary>
        /// Throw a thread and listen for communication
        /// </summary>        
        private void Listen()
        {
            vContinueListen = true;
            vClient = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            vServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            vServer.Bind(new IPEndPoint(this.ServerIP, this.ServerPort));
            vServer.Listen(1);

            for (; ; )
            {
                try
                {
                    //If there is not connection acceptation in one second => continue and come back
                    if (vServer.Poll(1000, SelectMode.SelectRead))
                        vClient = vServer.Accept();
                    //We have got a communication. Reading and throws an event
                    else if (vClient != null && vClient.Connected)
                    {
                        byte[] buffer = new byte[vClient.ReceiveBufferSize];
                        NetworkStream ns = new NetworkStream(vClient);
                        if (ns.DataAvailable)
                        {
                            IFormatter formatter = new BinaryFormatter();
                            this.Message = (object)formatter.Deserialize(ns);

                            #region Throw Event
                            NetworkThreadEventArgs arg = new NetworkThreadEventArgs();
                            arg.Message = this.Message;
                            OnReceivedMessage(arg);
                            #endregion
                        }
                    }
                    //End of the communication
                    else if (!vContinueListen)
                    {
                        if (vClient != null)
                            vClient.Close();
                        if (vServer != null)
                            vServer.Close();
                        return;
                    }
                }
                catch (ThreadAbortException) { }
                catch (Exception)
                {
#if DEBUG
                    throw;
#endif
                }
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the server IP
        /// </summary>
        public IPAddress ServerIP { get { return vServerIP; } set { vServerIP = value; } }
        /// <summary>
        /// Gets or sets the communication port of the server
        /// </summary>
        public int ServerPort { get { return vServerPort; } set { vServerPort = value; } }
        /// <summary>
        /// Gets the last received message from the network
        /// </summary>
        private object Message { get { return vMessage; } set { vMessage = value; } }
        #endregion

        #region Events
        /// <summary>
        /// Throwed for a received message from the network
        /// </summary>
        public event MessageReceivedEventHandler ReceivedMessage;
        /// <summary>
        /// Notify a received message
        /// </summary>
        /// <param name="args"></param>
        protected virtual void OnReceivedMessage(NetworkThreadEventArgs args)
        {
            if (ReceivedMessage != null)
                ReceivedMessage(this, args);
        }
        #endregion
    }
}
